-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore ALU.VHD 1.10
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
-- Philosophy    see: information/uCore_Philosophy.pdf
--
-- Arithmetic-Logic-Unit for pre-incrementing absolute data 
-- memory access.

LIBRARY IEEE;
USE     IEEE.std_logic_1164.ALL;
USE     IEEE.std_logic_signed.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;

ENTITY alu IS
PORT (reset     : IN    std_logic; 
      clk       : IN    std_logic; 
      clk_en    : IN    std_logic; 
      tri_en    : IN    std_logic;  
      inst      : IN    inst_bus;
      uBus      : IN    data_bus;
      tos       : IN    data_bus;
      lit       : IN    data_bus;
      nos       : IN    data_bus;
      pc        : IN    data_bus;
      rsp       : IN    data_bus;
      task      : IN    data_bus;
      stBus     : INOUT status_rec;
      memaddr   : OUT   data_bus;
      alu       : OUT   data_bus);
END alu;

ARCHITECTURE rtl OF alu IS

SIGNAL i_group    : inst_group;
SIGNAL i_stack    : inst_stack;
SIGNAL i_type     : inst_type;
SIGNAL i_lit      : std_logic;

SIGNAL adder_in1  : data_bus;
SIGNAL adder_mux  : data_bus;
SIGNAL result     : std_logic_vector(data_width DOWNTO 0);
SIGNAL sum        : data_bus;
SIGNAL alu_i      : data_bus;
SIGNAL carry      : std_logic;
SIGNAL carry_in   : std_logic;
SIGNAL carry_out  : std_logic;
SIGNAL s_c        : std_logic;

BEGIN

i_lit   <= inst(inst_width-1);
i_type  <= inst(inst_width-2 DOWNTO inst_width-3);
i_stack <= inst(inst_width-4 DOWNTO inst_width-5);
i_group <= inst(inst_width-6 DOWNTO 0);

memaddr <= alu_i;  -- pre incrementing memory access
alu     <= alu_i;

-- adder_in1 is tos and is prepared by the process alu_output
result  <= adder(adder_in1, adder_mux, carry_in);
sum     <= result(data_width-1 DOWNTO 0);
carry   <= result(data_width);

-----------------------------------------------------------------------
-- alu_output: 
-----------------------------------------------------------------------

alu_output: PROCESS(i_lit, i_type, i_stack, i_group, pc, carry, tos, nos, sum, s_c, rsp, task)
BEGIN

adder_in1 <= tos;
adder_mux <= pc;
alu_i     <= sum;
carry_in  <= '0';
carry_out <= s_c;

IF  i_lit='0'  THEN
   CASE  i_type  IS
     WHEN op_BRA => CASE i_stack IS
                      WHEN op_NONE => NULL;
                      WHEN op_POP  => NULL;
                      WHEN op_PUSH => NULL;
                      WHEN OTHERS  => NULL;
                    END CASE;
     WHEN op_ALU => CASE i_stack IS
                      WHEN op_BOTH => CASE  i_group  IS
                                        WHEN op_NOT  => alu_i <= NOT tos;
                                        WHEN op_SL   => alu_i <= tos(data_width-2 DOWNTO 0) & '0';  -- logical shift left: 0 goes into 0th bit pos.
                                                        carry_out  <= tos(data_width-1);          -- this can be used for asl too.
                                        WHEN op_ASR  => alu_i <= tos(data_width-1) & tos(data_width-1 DOWNTO 1); -- preserve the sign.
                                                        carry_out <= tos(0);
                                        WHEN op_LSR  => alu_i <= '0' & tos(data_width-1 DOWNTO 1);  -- LSR: 0 goes into MSB position.
                                                        carry_out <= tos(0);
                                        WHEN op_ROR  => alu_i <= s_c & tos(data_width-1 DOWNTO 1);  -- Rotation right through carry. Watch out range limits.
                                                        carry_out <= tos(0);
                                        WHEN op_ROL  => alu_i <= tos(data_width-2 DOWNTO 0) & s_c;
                                                        carry_out <= tos(data_width-1);
                                        WHEN op_ZEQU => IF  tos=slice('0', data_width)  THEN      -- Equal zero: State of tos.
                                                          alu_i <= (OTHERS => '1');
                                                        ELSE
                                                          alu_i <= (OTHERS => '0');
                                                        END IF;
                                        WHEN op_CC   => carry_out <= NOT s_c;
                                                        alu_i       <= tos;
                                        WHEN OTHERS  => NULL;
                                      END CASE;
                      WHEN OTHERS  => adder_mux <= nos;
                                      CASE  i_group  IS
                                        WHEN op_ADD => carry_out <= carry;
                                        WHEN op_ADC => carry_in  <= s_c;
                                                       carry_out <= carry;
                                        WHEN op_SUB => adder_in1 <= NOT tos;
                                                       carry_in  <= '1';
                                                       carry_out <= carry;
                                        WHEN op_SBC => adder_in1 <= NOT tos;
                                                       carry_in  <= s_c;
                                                       carry_out <= carry;
                                        WHEN op_AND => alu_i <= tos AND nos;
                                        WHEN op_OR  => alu_i <= tos OR  nos;
                                        WHEN op_XOR => alu_i <= tos XOR nos;
                                        WHEN OTHERS => alu_i <= nos;
                                      END CASE;
                      END CASE;
     WHEN op_MEM => CASE i_stack IS
                      WHEN op_NONE => CASE  i_group  IS
                                        WHEN op_LOCAL => IF  locals='1'  THEN
                                                            adder_mux <= rsp;
                                                         END IF;
                                        WHEN op_TASK  => IF  tasks='1'  THEN
                                                            adder_mux <= task;
                                                         END IF;
                                        WHEN OTHERS   => NULL;
                                      END CASE;
                      WHEN op_POP  => adder_mux <= slice(i_group(2), data_width-3) & i_group;
                      WHEN op_PUSH => adder_mux <= slice(i_group(2), data_width-3) & i_group;
                      WHEN OTHERS  => CASE  i_group  IS
                                        WHEN op_LOCAL => IF  locals='1'  THEN
                                                            adder_mux <= rsp;
                                                         END IF;
                                        WHEN op_TASK  => IF  tasks='1'  THEN
                                                            adder_mux <= task;
                                                         END IF;
                                        WHEN OTHERS   => NULL;
                                      END CASE;
                    END CASE;
     WHEN OTHERS => NULL;
   END CASE;   
END IF;

END PROCESS alu_output;

carry_status: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      s_c <= '0';
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1'  THEN
        s_c <= carry_out;
        IF  stBus.write='1'  THEN
           s_c <= stBus.tos(s_c_bit);
        END IF;
     END IF;
   END IF;
END PROCESS carry_status;

stBus.flags(s_c_bit) <= s_c;

END rtl;
