-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore FUNCTIONS.VHD 1.10
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
-- Philosophy    see: information/uCore_Philosophy.pdf
--
-- Functions and components which are used throughout
-- MicroCore.

LIBRARY IEEE;
USE     IEEE.std_logic_1164.ALL;
USE     IEEE.std_logic_arith.ALL;
USE     IEEE.std_logic_signed.ALL;

PACKAGE functions IS

FUNCTION  to_int(v : IN std_logic_vector ) RETURN INTEGER;

FUNCTION  to_vec(v : IN INTEGER;
                 s : IN INTEGER          ) RETURN std_logic_vector;

FUNCTION   slice(v : IN std_logic;
                 s : IN INTEGER          ) RETURN std_logic_vector;

FUNCTION     crc(d : IN std_logic_vector;
                 p : IN std_logic_vector ) RETURN std_logic_vector;

FUNCTION     max(v : IN INTEGER;
                 w : IN INTEGER          ) RETURN INTEGER;

FUNCTION    log2(v : IN INTEGER          ) RETURN INTEGER;

FUNCTION   adder(v : IN std_logic_vector;
                 w : IN std_logic_vector;
                 carry : IN std_logic    ) RETURN std_logic_vector;

END functions;


PACKAGE BODY functions IS

    ----------------------------------------------------------------
    -- to_int
    -- Convert std_logic_vector to integer number 
    ----------------------------------------------------------------

    FUNCTION to_int(v  :  IN std_logic_vector
                   )  RETURN INTEGER IS
        VARIABLE temp  :     INTEGER;
    BEGIN
        temp := conv_integer(v);
        RETURN temp;
    END;

    ----------------------------------------------------------------
    -- to_vec
    -- Convert integer number to std_logic_vector
    ----------------------------------------------------------------

    FUNCTION to_vec(v  :  IN INTEGER;
                    s  :  IN INTEGER
                   )  RETURN std_logic_vector IS
        VARIABLE temp  :     std_logic_vector(s - 1 DOWNTO 0);
    BEGIN
        temp := conv_std_logic_vector(v,s);
        RETURN temp;
    END;

    ----------------------------------------------------------------
    -- slice 
    -- Return (others => v) within range s-1 to 0
    ----------------------------------------------------------------

    FUNCTION slice(v   :  IN std_logic;
                   s   :  IN INTEGER
                  )   RETURN std_logic_vector IS
        VARIABLE temp  :     std_logic_vector (s - 1  DOWNTO 0);
    BEGIN
        temp := (OTHERS => v); -- synopsys compatibility
        RETURN temp;
    END;

    ----------------------------------------------------------------
    -- crc-function (not for synthesis)
    ----------------------------------------------------------------

    FUNCTION crc(d     :  IN std_logic_vector;
                 p     :  IN std_logic_vector
                )     RETURN std_logic_vector IS
        VARIABLE temp  :     std_logic_vector (p'HIGH DOWNTO p'LOW);
    BEGIN
        temp := (OTHERS => '0');
        FOR I IN d'HIGH DOWNTO  d'LOW LOOP
            IF   d(I) = temp(p'HIGH) THEN
                temp := temp(p'HIGH - 1 DOWNTO p'LOW) & '0';
            ELSE
                temp := temp(p'HIGH - 1 DOWNTO p'LOW) & '0';
                temp := temp XOR p;
            END IF;
        END LOOP;
        RETURN temp;
    END;

    ----------------------------------------------------------------
    -- max
    -- Return maximum of two integers w and v
    ----------------------------------------------------------------

    FUNCTION max(v  :  IN INTEGER;
                 w  :  IN INTEGER
                )  RETURN INTEGER IS
    BEGIN
        IF (v > w) THEN  RETURN v;  ELSE  RETURN w;  END IF;
    END;

    ----------------------------------------------------------------
    -- logarithm dualis
    ----------------------------------------------------------------

    FUNCTION log2(v    :  IN INTEGER
                 )    RETURN INTEGER IS
        VARIABLE temp  :     INTEGER;
    BEGIN
        temp := 0;
        WHILE  2 ** temp < v  LOOP
            temp := temp + 1;
        END LOOP;
        RETURN temp;
    END;

    ----------------------------------------------------------------
    -- adder of the alu
    ----------------------------------------------------------------

    FUNCTION adder(v     : IN  std_logic_vector;
                   w     : IN  std_logic_vector;
                   carry : IN  std_logic
                  ) RETURN std_logic_vector IS
        VARIABLE temp : std_logic_vector (v'HIGH + 1 DOWNTO v'LOW);
    BEGIN
        temp := ('0' & v) + ('0' & w) + carry;
        RETURN temp;
    END;

END functions;


----------------------------------------------------------------
-- synchronising asynchronous input signals
----------------------------------------------------------------

LIBRARY IEEE;
USE     IEEE.std_logic_1164.ALL;

ENTITY synchronise IS
PORT (reset    : IN    std_logic;
      clk      : IN    std_logic;

      async    : IN    std_logic; 
      sync     : OUT   std_logic);
END synchronise;

ARCHITECTURE rtl OF synchronise IS
BEGIN

synch: PROCESS(clk, reset)
BEGIN
  IF  reset='1'  THEN
    sync <= '0';
  ELSIF  rising_edge(clk)  THEN
    sync <= async;
  END IF;
END PROCESS synch; 

END rtl;


----------------------------------------------------------------
-- synchronising inverted asynchronous input signals
----------------------------------------------------------------

LIBRARY IEEE;
USE     IEEE.std_logic_1164.ALL;

ENTITY synchronise_n IS
PORT (reset    : IN    std_logic;
      clk      : IN    std_logic;

      async_n  : IN    std_logic; 
      sync     : OUT   std_logic);
END synchronise_n;

ARCHITECTURE rtl OF synchronise_n IS
BEGIN

synch_n: PROCESS(clk, reset)
BEGIN
  IF  reset='1'  THEN
    sync <= '0';
  ELSIF  rising_edge(clk)  THEN
    sync <= NOT async_n;
  END IF;
END PROCESS synch_n; 

END rtl;