-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore IO.VHD 1.10
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
-- Philosophy    see: information/uCore_Philosophy.pdf
--
-- Interrupt signal preprocessing with FLAGS register
-- to signal interrupt sources and the IE register to
-- enable individual interrupts.

LIBRARY IEEE;
USE     IEEE.std_logic_1164.ALL;
USE     IEEE.std_logic_signed.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;

ENTITY io IS
PORT (reset     : IN    std_logic; 
      clk       : IN    std_logic; 
      clk_en    : IN    std_logic; 
      inst      : IN    inst_bus;
      memaddr   : IN    data_bus;
      nos       : IN    data_bus;
      ioBus     : INOUT io_port;
      task      : OUT   data_bus);
END io;

ARCHITECTURE rtl OF io IS

SIGNAL i_lit       : std_logic;
SIGNAL i_type      : inst_type;
SIGNAL i_stack     : inst_stack;
SIGNAL i_group     : inst_group;

SIGNAL iospace     : std_logic;
SIGNAL internal    : std_logic;
SIGNAL internal_wr : std_logic;
SIGNAL io_sel      : std_logic;
SIGNAL io_wr       : std_logic;
SIGNAL task_sel    : std_logic;

BEGIN

i_lit   <= inst(inst_width-1);
i_type  <= inst(inst_width-2 DOWNTO inst_width-3);
i_stack <= inst(inst_width-4 DOWNTO inst_width-5);
i_group <= inst(inst_width-6 DOWNTO 0);

-- ************** internally memory mapped registers *************************
-- internal is set when an internal register is accessed via memory mapped i/o
--
internal <= task_sel;

task_sel <= '1'  WHEN  memaddr=TASK_REGISTER AND tasks='1'  ELSE '0';

task_register: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      task <= (OTHERS => '0');
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1' AND internal_wr='1'  THEN
        IF  task_sel='1'  THEN
           task <= slice('0', data_width-data_addr_width) & nos(data_addr_width-1 DOWNTO 0);
        END IF;
     END IF;
   END IF;
END PROCESS task_register;


-- ****************************************************************************
--
-- external ioBus control
--

ioBus.ioaddr   <= memaddr(data_width-1);
ioBus.internal <= internal;
ioBus.sel      <= io_sel AND NOT internal;
ioBus.write    <= io_wr;
ioBus.addr     <= memaddr(data_addr_width-1 DOWNTO 0);
ioBus.dout     <= nos;
iospace        <= memaddr(data_width-1) AND NOT internal;

io_control: PROCESS (i_lit, i_type, i_stack, iospace, nos, memaddr)
BEGIN

io_sel <= '0';
io_wr <= '0';
internal_wr <= '0';

IF  i_lit='0'  THEN
   CASE  i_type  IS
     WHEN op_MEM => CASE i_stack IS
                      WHEN op_POP  => IF  iospace='1'  THEN
                                         io_sel <= '1';
                                         io_wr <= '1';
                                      ELSE
                                         internal_wr <= '1';
                                      END IF;
                      WHEN op_PUSH => IF  iospace='1'  THEN
                                         io_sel <= '1';
                                      END IF;                       
                      WHEN OTHERS  => NULL;
                    END CASE;
     WHEN OTHERS => NULL;
   END CASE;   
END IF;
END PROCESS io_control;

END rtl;
