-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore RSTACK.VHD 1.10
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
-- Philosophy    see: information/uCore_Philosophy.pdf
--
-- The return stack MicroCell. It uses the topmost data
-- ram area and and can be addressed using absolute or return
-- stack pointer relative addresses when LOCALS is true.

LIBRARY IEEE;
USE     IEEE.std_logic_1164.ALL;
USE     IEEE.std_logic_signed.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;

ENTITY r_stack IS
PORT (reset     : IN    std_logic; 
      clk       : IN    std_logic; 
      clk_en    : IN    std_logic; 
      inst      : IN    inst_bus;
      condition : IN    std_logic;
      uBus      : IN    data_bus;
      ioBus     : IN    io_port;
      memaddr   : IN    data_bus;
      nos       : IN    data_bus;
      alu       : IN    data_bus;
      pc        : IN    data_bus;
      dram      : INOUT dram_port;
      rsp       : OUT   data_bus);
END r_stack;

ARCHITECTURE rtl OF r_stack IS

SIGNAL i_lit      : std_logic;
SIGNAL i_type     : inst_type;
SIGNAL i_stack    : inst_stack;
SIGNAL i_group    : inst_group;

SIGNAL ioaddr     : std_logic;
SIGNAL din        : data_bus;
SIGNAL dout       : data_bus;
SIGNAL addr       : data_addr;
SIGNAL write      : std_logic;

SIGNAL rsp_reg    : rs_addr;
SIGNAL rsp_new    : rs_addr;

BEGIN

i_lit   <= inst(inst_width-1);
i_type  <= inst(inst_width-2 DOWNTO inst_width-3);
i_stack <= inst(inst_width-4 DOWNTO inst_width-5);
i_group <= inst(inst_width-6 DOWNTO 0);
ioaddr  <= ioBus.ioaddr;

rsp <= '0' & slice('1', data_width-rs_addr_width-1) & rsp_reg;

dram.write  <= write;   
dram.addr   <= addr; 
dram.din    <= din; 

------------------------------------------------------------------------------
-- return stack grows from higher to lower addresses
------------------------------------------------------------------------------
rstack_control: PROCESS (rsp_reg, pc, memaddr, i_lit, i_type, i_stack, i_group, condition, ioaddr, uBus, alu)

PROCEDURE push_stack IS
BEGIN
  rsp_new <= rsp_reg-1;
  write <= '1';
  addr <= slice('1', data_addr_width - rs_addr_width) & rsp_reg-1;
END push_stack;

PROCEDURE pop_stack IS
BEGIN
  rsp_new  <= rsp_reg+1;
  write <= '0';
  addr <= slice('1', data_addr_width - rs_addr_width) & rsp_reg;
END pop_stack;

BEGIN

din <= pc;
addr <= memaddr(data_addr_width-1 DOWNTO 0);
rsp_new <= rsp_reg;
write <= '0';

IF  i_lit='0'  THEN
   CASE  i_type  IS
     WHEN op_BRA => CASE i_stack IS
                      WHEN op_NONE => IF  condition='1'  THEN
                                         din <= uBus;
                                         pop_stack;
                                      END IF;
                      WHEN op_POP  => NULL;
                      WHEN op_PUSH => NULL;
                      WHEN OTHERS  => IF  condition='1'  THEN
                                         push_stack;
                                      END IF;
                    END CASE;
     WHEN op_ALU => NULL;
     WHEN op_MEM => CASE i_stack IS
                      WHEN op_NONE => CASE  i_group  IS
                                        WHEN op_RSTACK => din <= uBus;
                                                          push_stack;
                                        WHEN op_RSP    => rsp_new <= uBus(rs_addr_width-1 DOWNTO 0);
                                        WHEN OTHERS    => IF  locals='1' AND i_group=op_LOCAL  THEN   -- otherwise, the synthesiser does not
                                                             addr <= alu(data_addr_width-1 DOWNTO 0); -- handle locals='0' gracefully
                                                             din <= uBus;
                                                             write <= '1';
                                                          END IF;
                                                          IF  tasks='1' AND i_group=op_TASK  THEN
                                                             addr <= alu(data_addr_width-1 DOWNTO 0);
                                                             din <= uBus;
                                                             write <= '1';
                                                          END IF;
                                      END CASE;
                      WHEN op_POP  => din <= uBus;                                     
                                      IF  ioaddr='0'  THEN
                                          write <= '1';
                                      END IF;
                      WHEN op_PUSH => NULL;                     
                      WHEN OTHERS  => CASE  i_group  IS
                                        WHEN op_RSTACK => pop_stack;
                                        WHEN op_TOR    => addr <= slice('1', data_addr_width - rs_addr_width) & rsp_reg;
                                        WHEN OTHERS    => IF  locals='1' AND i_group=op_LOCAL  THEN    -- otherwise, the synthesiser does not
                                                             addr <= alu(data_addr_width-1 DOWNTO 0);  -- handle locals='0' gracefully
                                                          END IF;
                                                          IF  tasks='1'  AND i_group=op_TASK  THEN
                                                             addr <= alu(data_addr_width-1 DOWNTO 0);
                                                          END IF;
                                      END CASE;
                    END CASE;
     WHEN OTHERS => NULL;
   END CASE;   
END IF;
END PROCESS rstack_control;

rsp_register: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      rsp_reg <= (OTHERS => '1');
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1'  THEN
        rsp_reg <= rsp_new;
     END IF;
   END IF;
END PROCESS rsp_register;

END rtl;
