-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore SEQUENCER.VHD 1.10
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
-- Philosophy    see: information/uCore_Philosophy.pdf
--
-- The sequencer generates the address for the next
-- instruction fetch from program ROM. It realises
-- interrupt processing and the trap mechanism for
-- multitasking.

LIBRARY IEEE;
USE     IEEE.std_logic_1164.ALL;
USE     IEEE.std_logic_signed.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;

ENTITY sequencer IS
PORT (reset     : IN    std_logic; 
      clk       : IN    std_logic; 
      clk_en    : IN    std_logic; 
      int       : IN    std_logic;
      ioBus     : IN    io_port;
      stBus     : IN    status_rec;
      uBus      : IN    data_bus;
      inst      : OUT   inst_bus;
      condition : OUT   std_logic;
      pc        : OUT   data_bus;
      rom       : INOUT rom_port);
END sequencer;

ARCHITECTURE rtl OF sequencer IS

SIGNAL inst_reg   : inst_bus;
SIGNAL i_lit      : std_logic;
SIGNAL i_type     : inst_type;
SIGNAL i_stack    : inst_stack;
SIGNAL i_group    : inst_group;
SIGNAL pc_reg     : prog_addr;
SIGNAL addr       : prog_addr;
SIGNAL cond_i     : std_logic;
SIGNAL trap       : std_logic;

BEGIN

inst      <= inst_reg;
rom.addr  <= addr;
condition <= cond_i;
i_lit     <= inst_reg(inst_width-1);
i_type    <= inst_reg(inst_width-2 DOWNTO inst_width-3);
i_stack   <= inst_reg(inst_width-4 DOWNTO inst_width-5);
i_group   <= inst_reg(inst_width-6 DOWNTO 0);
trap      <= ioBus.trap;

pc <= slice('0', data_width-prog_addr_width) & pc_reg;

------------------------------------------------------------------------------
-- sequencer_control: Generates control lines for the sequencer entity
-- The opcodes' fields are defined in instruction register as follows:
--   7   6   5   4   3   2   1   0
-- |---|---|---|---|---|---|---|---|
-- |   |   |   |   |   |   |   |   |
-- |---|---|---|---|---|---|---|---|
-- |LIT|  TYPE | STACK |    GROUP  |
-- |---|-------|-------|-----------|
--
-- In the process the op-field are preceded by "i_" .
-- The different "op_XXX" are defined in constants.vhdl.
-- Remember the condition signal is defined as follows :
--      condition  =   s_z      WHEN i_group = op_ZERO,
--                 =   NOT s_z  WHEN i_group = op_NZERO,
--                 =   NOT s_n  WHEN i_group = op_NSIGN,
--                 =   NOT s_c  WHEN i_group = op_NCARRY,
--                 =   '0'      WHEN i_group = op_NEVER,
--                 =   '1'      WHEN i_group = OTHERS; -- op_PAUSE | op_INT | op_ALWAYS,
-- 
------------------------------------------------------------------------------

WITH  i_group  SELECT
cond_i <=  stBus.flags(s_z_bit)  WHEN op_ZERO,
       NOT stBus.flags(s_z_bit)  WHEN op_NZERO,
       NOT stBus.flags(s_n_bit)  WHEN op_NSIGN,
       NOT stBus.flags(s_c_bit)  WHEN op_NCARRY,
                            '0'  WHEN op_NEVER,
                            '1'  WHEN OTHERS;  --op_PAUSE | op_INT | op_ALWAYS,

------------------------------------------------------------------------------
-- Build program's memory address. 
-- You have the following choices:
-- rom address = pc_reg
-- rom address = uBus
-- rom address = 0x0004 address of interrupt service routine
-- rom address = 0x0008 address of task service routine
------------------------------------------------------------------------------

addr_select: PROCESS (i_lit, i_type, i_stack, i_group, cond_i, pc_reg, uBus)
BEGIN

addr <= pc_reg;

IF  i_lit='0' AND cond_i='1'  THEN
   CASE  i_type  IS
     WHEN op_BRA => addr <= uBus(prog_addr_width-1 DOWNTO 0);
                    CASE i_stack IS
                      WHEN op_BOTH => IF  i_group=op_PAUSE  THEN
                                         addr <= slice('0', prog_addr_width-4) & addr_tsr;
                                      ELSIF  i_group=op_INT  THEN
                                         addr <= slice('0', prog_addr_width-4) & addr_isr;
                                      END IF;
                      WHEN OTHERS  => NULL;
                    END CASE;
     WHEN OTHERS => NULL;
   END CASE;
END IF;

END PROCESS addr_select;

-----------------------------------------------------------------------
-- pc_register (program counter)
-- (a) reset = 1 : pc_reg = 0 
-- (b) clk valid : pc_reg = addr+1
-- (c) clk valid & int = 1: pc_reg = addr (no increment during an interrupt acknowlege)
------------------------------------------------------------------------
pc_register: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      pc_reg <= (OTHERS => '0');
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1'  THEN
       pc_reg <= addr + 1;
       IF  int='1'  THEN
          pc_reg <= addr;
       END IF;
     END IF;
   END IF;
END PROCESS pc_register;

-----------------------------------------------------------------------
-- inst_register (instruction register)
-- (a) reset = 1:  inst = NOP 
-- (b) clk valid:  inst = rom.data
-- (c) clk valid & int  = 1: inst = INT_OP    (BRA BOTH INT)
-- (d) clk valid & trap = 1: inst = PAUSE_OP  (BRA BOTH PAUSE)
-----------------------------------------------------------------------
inst_register: PROCESS(clk, reset)
BEGIN
   IF  reset='1'  THEN
      inst_reg <= NO_OP;
   ELSIF  rising_edge(clk)  THEN
     IF  clk_en='1' OR trap='1'  THEN
       inst_reg <= rom.data;
       IF  int='1'  THEN
          inst_reg <= INT_OP;
       END IF;
       IF  trap='1'  THEN
          inst_reg <= PAUSE_OP;
       END IF;
     END IF;
   END IF;
END PROCESS inst_register;

END rtl;
