-- Do not use this file except in compliance with the License. You may
-- obtain a copy of the License at http://www.microcore.org/License/
-- Software distributed under the License is distributed on an "AS IS" basis,
-- WITHOUT WARRANTY OF ANY KIND, either express or implied.
-- See the License for the specific language governing rights and limitations
-- under the License.
--
-- The Original Code is: MicroCore UBUS.VHD 1.10
--
-- The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
--
-- Introduction  see: information/uCore_Introduction.pdf
-- Tools         see: information/uCore_Implementation.pdf
-- Versions      see: uCore_Versions.txt
-- License       see: information/uCore_License.pdf
-- Conventions   see: information/uCore_Conventions.pdf
-- Contributors  see: information/uCore_Contributors.pdf
-- Philosophy    see: information/uCore_Philosophy.pdf
--
-- Internal bus based on a hierarchy of multiplexers. This file 
-- should only be used when the technology does not support 
-- 3-state buffers. Otherwise, use uBus3state.vhd, which
-- implements the bus using 3-state buffers.

LIBRARY IEEE;
USE     IEEE.std_logic_1164.ALL;
USE     work.functions.ALL;
USE     work.constants.ALL;

ENTITY microbus IS
PORT (reset     : IN    std_logic; 
      clk       : IN    std_logic; 
      clk_en    : IN    std_logic;
      tri_en    : IN    std_logic;  
      inst      : IN    inst_bus;
      stBus     : IN    status_rec;
      tos       : IN    data_bus;
      nos       : IN    data_bus;
      alu       : IN    data_bus;
      dsp       : IN    data_bus;
      rsp       : IN    data_bus;
      dram      : IN    dram_port;
      flags     : IN    data_bus;
      task      : IN    data_bus;
      ioBus     : IN    io_port;
      uBus      : INOUT data_bus);
END microbus;

ARCHITECTURE multiplexed OF microbus IS

ATTRIBUTE syn_keep : BOOLEAN;

SIGNAL i_lit      : std_logic;
SIGNAL i_type     : inst_type;
SIGNAL i_stack    : inst_stack;
SIGNAL i_group    : inst_group;
SIGNAL s_lit      : std_logic;
SIGNAL internal   : std_logic;
SIGNAL iosel      : std_logic;

SIGNAL data       : data_bus;  ATTRIBUTE syn_keep OF data : SIGNAL IS true;  -- forces multiplexer hierarchy
SIGNAL regs       : data_bus;  ATTRIBUTE syn_keep OF regs : SIGNAL IS true;  -- forces multiplexer hierarchy

SIGNAL ram_sel    : std_logic;
SIGNAL status_sel : std_logic;
SIGNAL rsp_sel    : std_logic; 
SIGNAL dsp_sel    : std_logic;
SIGNAL flags_sel  : std_logic;
SIGNAL tos_sel    : std_logic;
SIGNAL nos_sel    : std_logic;
SIGNAL alu_sel    : std_logic;
SIGNAL io_sel     : std_logic;
SIGNAL task_sel   : std_logic; 

BEGIN

i_lit   <= inst(inst_width-1);
i_type  <= inst(inst_width-2 DOWNTO inst_width-3);
i_stack <= inst(inst_width-4 DOWNTO inst_width-5);
i_group <= inst(inst_width-6 DOWNTO 0);
s_lit   <= stBus.flags(s_lit_bit);

internal <= ioBus.internal;
iosel    <= ioBus.sel;

uBus <= data  WHEN  (tos_sel OR nos_sel OR io_sel OR ram_sel)='1'
   ELSE regs  WHEN  (dsp_sel OR rsp_sel OR flags_sel OR status_sel)='1'
   ELSE task  WHEN  (task_sel AND tasks)='1'
   ELSE alu ;

data <= tos       WHEN  tos_sel='1' 
   ELSE nos       WHEN  nos_sel='1'
   ELSE ioBus.din WHEN  io_sel ='1'
   ELSE dram.dout ;  -- ram_sel='1'

regs <= dsp   WHEN  dsp_sel='1'
   ELSE rsp   WHEN  rsp_sel='1'
   ELSE flags WHEN  flags_sel='1'
   ELSE slice('0',data_width - status_width) & stBus.flags ; -- status_sel='1'



uBus_control: PROCESS (i_lit, i_type, i_stack, i_group, s_lit, iosel, internal)

PROCEDURE activate (SIGNAL o : OUT std_logic) IS
BEGIN
  o <= '1';
  alu_sel <= '0';
END activate;

BEGIN

alu_sel    <= '1';
ram_sel    <= '0';
status_sel <= '0';
rsp_sel    <= '0';
dsp_sel    <= '0';
flags_sel  <= '0';
tos_sel    <= '0';
nos_sel    <= '0';
io_sel     <= '0';
task_sel   <= '0';

IF  i_lit='0'  THEN
   CASE  i_type  IS
     WHEN op_BRA => CASE i_stack IS
                      WHEN op_NONE => activate(ram_sel);
                      WHEN op_POP  => IF  s_lit='0'  THEN
                                         activate(tos_sel);
                                      END IF;
                      WHEN op_PUSH => NULL;
                      WHEN OTHERS  => IF  i_group=op_INT OR i_group=op_PAUSE  THEN
                                         activate(status_sel);
                                      ELSIF  s_lit='0'  THEN
                                         activate(tos_sel);
                                      END IF;
                      END CASE;
     WHEN op_ALU => NULL;
     WHEN op_MEM => CASE i_stack IS
                      WHEN op_NONE => CASE  i_group  IS
                                        WHEN op_LOCAL => IF  locals='1'  THEN
                                                            activate(nos_sel);
                                                         ELSE
                                                            activate(tos_sel);
                                                         END IF;
                                        WHEN op_TASK  => IF  tasks='1'  THEN
                                                            activate(nos_sel);
                                                         ELSE
                                                            activate(tos_sel);                                        
                                                         END IF;
                                        WHEN OTHERS   => activate(tos_sel);
                                      END CASE;
                      WHEN op_POP  => activate(nos_sel);
                      WHEN op_PUSH => IF  iosel='1'  THEN
                                          activate(io_sel);
                                      ELSIF  internal='1' AND tasks='1'  THEN
                                          activate(task_sel);
                                      ELSE
                                          activate(ram_sel);
                                      END IF;                       
                      WHEN OTHERS  => CASE i_group IS
                                        WHEN op_STATUS => activate(status_sel);
                                        WHEN op_TOR    => activate(ram_sel);
                                        WHEN op_RSTACK => activate(ram_sel);
                                        WHEN op_LOCAL  => IF  locals='1'  THEN
                                                             activate(ram_sel);
                                                          END IF;
                                        WHEN op_RSP    => activate(rsp_sel);
                                        WHEN op_DSP    => activate(dsp_sel);
                                        WHEN op_TASK   => IF  tasks='1'  THEN
                                                             activate(ram_sel);
                                                          END IF;
                                        WHEN OTHERS    => activate(flags_sel); -- op_FLAGS | op_IE
                                      END CASE;
                    END CASE;
     WHEN OTHERS => NULL;
   END CASE;
END IF;
END PROCESS uBus_control;

END multiplexed;
